/*
 *	This file is part of k7mctrl
 *      Copyright(C) 2004 NAKAYAMA Hiroyoshi
 *
 * 	Licensed under the terms of the GNU GPL License version 2.
 */
#include <stdio.h>
#include "miscio.h"

#include "devmsr.h"

int FID_DATA[32] = {
  110, 115, 120, 125,  50,  55,  60,  65,
   70,  75,  80,  85,  90,  95, 100, 105,
   30, 190,  40, 200, 130, 135, 140, 210,
  150, 220, 160, 165, 170, 180, 230, 240,
};

float VID_DATA_DESKTOP[32] = {
1.850, 1.825, 1.800, 1.775, 1.750, 1.725, 1.700, 1.675, 1.650, 1.625,
1.600, 1.575, 1.550, 1.525, 1.500, 1.475, 1.450, 1.425, 1.400, 1.375,
1.350, 1.325, 1.300, 1.275, 1.250, 1.225, 1.200, 1.175, 1.150, 1.125,
1.100, 0,
};

float VID_DATA_MOBILE[32] = {
2.000, 1.950, 1.900, 1.850, 1.800, 1.750, 1.700, 1.650, 1.600, 1.550,
1.500, 1.450, 1.400, 1.350, 1.300, 0, 1.275, 1.250, 1.225, 1.200,
1.175, 1.150, 1.125, 1.100, 1.075, 1.050, 1.025, 1.000, 0.975, 0.950,
0.925, 0,
};

static void change_FID(int fid, unsigned int latency)
{
  union msr_fidvidctl fidvidctl;

  if (latency<100)
    latency = 100;

/*  printf("[change_FID]\n%02X, %d\n",fid,latency); */
  latency *= 100;

  rddevmsr (0, MSR_K7_FID_VID_CTL, &fidvidctl.val);
  if (fidvidctl.bits.FID != fid) {
    fidvidctl.bits.SGTC = latency;
    fidvidctl.bits.FID  = fid;
    fidvidctl.bits.VIDC = 0;
    fidvidctl.bits.FIDC = 1;

#if 0
    {
      unsigned long l,h;
      l = fidvidctl.val;
      h = fidvidctl.val>>32;
      printf("WRITING....%08X_%08X...%02X(%02X)\n",h,l,fidvidctl.bits.FID,fid);
    }

    printf("[[%d,%02X,%d,%d]]\n",
	   fidvidctl.bits.SGTC,
	   fidvidctl.bits.FID ,
	   fidvidctl.bits.VIDC,
	   fidvidctl.bits.FIDC);
#endif

    wrdevmsr (0, MSR_K7_FID_VID_CTL, &fidvidctl.val);
  }
}


static void change_VID(int vid, unsigned int latency)
{
  union msr_fidvidctl fidvidctl;

  if (latency<100)
    latency = 100;
  latency *= 100;

  rddevmsr (0, MSR_K7_FID_VID_CTL, &fidvidctl.val);
  if (fidvidctl.bits.VID != vid) {
    fidvidctl.bits.SGTC = latency;
    fidvidctl.bits.VID = vid;
    fidvidctl.bits.FIDC = 0;
    fidvidctl.bits.VIDC = 1;
    wrdevmsr (0, MSR_K7_FID_VID_CTL, &fidvidctl.val);
  }
}

static void check_FID_VID (void)
{
  union msr_fidvidstatus fvstatus;
  unsigned long l,h;

  if (rddevmsr (0, MSR_K7_FID_VID_STATUS, &fvstatus.val) != -1) {
    l = fvstatus.val;
    h = fvstatus.val>>32;
/*    printf("VAL=%08X %08X\n",h,l); */
    printf("Max     VID: %02X(%1.3fV, %1.3fV)\n",
	   fvstatus.bits.MVID
	   ,VID_DATA_DESKTOP[fvstatus.bits.MVID]
	   ,VID_DATA_MOBILE[fvstatus.bits.MVID]);
    printf("Start   VID: %02X(%1.3fV, %1.3fV)\n",
	   fvstatus.bits.SVID
	   ,VID_DATA_DESKTOP[fvstatus.bits.SVID]
	   ,VID_DATA_MOBILE[fvstatus.bits.SVID]);
    printf("Current VID: %02X(%1.3fV, %1.3fV)\n",
	   fvstatus.bits.CVID
	   ,VID_DATA_DESKTOP[fvstatus.bits.CVID]
	   ,VID_DATA_MOBILE[fvstatus.bits.CVID]);

    printf("Max     FID: %02X(%2.1fx) \n",
	   fvstatus.bits.MFID,
	   (float)FID_DATA[fvstatus.bits.MFID]/10);
    printf("Start   FID: %02X(%2.1fx) \n",
	   fvstatus.bits.SFID,
	   (float)FID_DATA[fvstatus.bits.SFID]/10);
    printf("Current FID: %02X(%2.1fx) \n",
	   fvstatus.bits.CFID,
	   (float)FID_DATA[fvstatus.bits.CFID]/10);
  }
}

static struct ARGOPTION opt[] = { 
  { "-fid",    FLOAT,   0},	/* FID */
  { "-vid",    FLOAT,   0},	/* vid */
  { "-help",     FLAG,    0}, 
  { "--help",    FLAG,    0}, 
  { "-h",        FLAG,    0}, 
  { "",           END,    0}, 
};

void help(int argc, char **argv)
{
  printf("usage: %s [option]\n",*argv); /* Command Name. */
  printf(" no option : show status.\n");
  printf(" -fid f (f=3.0 ... 24.0)   : change fid\n");
  printf(" -vid v (v=0.925 ... 2.00) : change vid\n");
  printf(" -help | --help | -h : show this help.\n");
}

int main(int argc, char **argv)
{
  float fid = 0, vid = 0;
  int i;

  ArgGetOption(opt, argc, argv); 
  i = 0;
  while (opt[i].type != END ) {
    if (!strcmp("-fid",opt[i].option)) { 
      if (opt[i].data.fdata > 0) 
	fid = (opt[i].data.fdata); 
    } 
    if (!strcmp("-vid",opt[i].option)) {
      if (opt[i].data.fdata > 0) 
	vid = (opt[i].data.fdata); 
    } 
    if (!strcmp("-h",opt[i].option) ||
	!strcmp("-help",opt[i].option) ||
	!strcmp("--help",opt[i].option)) {
      if (opt[i].data.flag) {
	help(argc, argv);
	exit(0);
      }
    } 
    i ++; 
  }

  if (iopl(3)!=0) {
    printf("Not enough permissions.\n");
    exit(1);
  }

  /* printf("FID:%2.1f VID:%1.3f\n",fid,vid);  */

  if (fid > 0 ) {
    int j;
    for (i = 0 ; i < 32; i ++) {
      if ((float)FID_DATA[i]/10 == fid) {
	printf("FID: %02X(%2.1fx)\n",i,(float)FID_DATA[i]/10);
	j = i;
      }
    }
    change_FID(j,100);
  }

  if (fid == 0 && vid > 0) {
    int j;
    for (i = 0 ; i < 32; i ++) {
      if (VID_DATA_MOBILE[i] == vid) {
	printf("VID: %02X(%1.3fV, %1.3fV)\n",i,VID_DATA_DESKTOP[i],VID_DATA_MOBILE[i]);
	j = i;
      }
    }
    change_VID(j,100);
  }

  if (fid == 0 && vid == 0)
    check_FID_VID();
}
