/*
 *      Copyright(C) 2004 NAKAYAMA Hiroyoshi
 *
 * 	Licensed under the terms of the GNU GPL License version 2.
 */
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/io.h>

#define __USE_UNIX98	/* for pread/pwrite */
#define __USE_FILE_OFFSET64 /* we should use 64 bit offset for pread/pwrite */
#include <unistd.h>

#include <libintl.h>

#include "devmsr.h"

/* #define DEBUG */

int rddevmsr(int cpu, unsigned long index, unsigned long long *val)
{
  char cpuname[16];
  int fh;
  ssize_t r_read;
  
  snprintf (cpuname, 15,"/dev/cpu/%d/msr", cpu);
  fh = open (cpuname, O_RDONLY);
#ifdef DEBUG
  printf("[ rdmsr  start ]\n%s (fh = %d)\n",cpuname, fh);
#endif
  
  if ( fh == -1 ) {
    printf("rddevmsr:Not such a device(%s).\n",cpuname );
    return(-1);
  }

  r_read = pread(fh, val, 8, index);

#ifdef DEBUG
  {
    unsigned long l,h;
    l = *val;
    h = (*val)>>32;
    printf("pread index=%08X , val=%08X_%08X\n",index,h,l);
  }
#endif
  close (fh);
  
#ifdef DEBUG
  printf("[ rdmsr  end (%d)]\n",r_read);
#endif

  return(r_read);
}

int wrdevmsr(int cpu, unsigned long index, unsigned long long *val)
{
  int fh;
  char cpuname[16];
  int r;

  snprintf (cpuname,15, "/dev/cpu/%d/msr", cpu);
  fh = open (cpuname, O_WRONLY);

#ifdef DEBUG
  {
    unsigned long l,h;

    printf("func:wrmsr()\n");
    printf(" open device \"%s\" at O_WRONLY\n",cpuname);
    l = *val;
    h = (*val)>>32;
    printf(" pwrite to %0xh, write value:%08x_%08xh(64bits)\n",index,h,l);
  }
#endif

  if (fh==-1) {
    printf("wrdevmsr:Not such a device(%s).\n",cpuname );
    return(-1);
  }

  r = pwrite(fh, val, 8, index);

  close (fh);

  return (r);
}
