/*
 *	This file is part of tkK7mauto
 *      Copyright(C) 2004 NAKAYAMA Hiroyoshi
 *
 * 	Licensed under the terms of the GNU GPL License version 2.
 */
#include <tcl.h>
#include <tk.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include "cpustat.h"
#include "maketable.h"
#include "k7mctrl.h"
#include "sensors.h"
#include "miscio.h"

static float fsb;
static unsigned char FIDtable[8];
static float FID_DATA[32] = {
  11.0, 11.5, 12.0, 12.5,  5.0,  5.5,  6.0,  6.5,
   7.0,  7.5,  8.0,  8.5,  9.0,  9.5, 10.0, 10.5,
   3.0, 19.0,  4.0, 20.0, 13.0, 13.5, 14.0, 21.0,
  15.0, 22.0, 16.0, 16.5, 17.0, 18.0, 23.0, 24.0,
};


#define STATMAXNUM 60
double cpustates[STATMAXNUM];
double cpuclocks[STATMAXNUM];
int time_index;
int nowclock_level;
int autof;

int f_minimam;
int f_maximam;

/* Tcl/Tk */
#define TEXEC( what, errstr ) if ( (what) != TCL_OK ) {\
   fprintf( stderr, errstr": %s\n", interp->result );\
   exit( 1 );\
};
/* Tcl/Tk */
#define TEXEC2( what, errstr ) (what); if ( *interp->result != 0 ) {\
   fprintf( stderr, errstr": %s\n", interp->result );\
   exit( 1 );\
};

static char *mix_init;		/* Tcl/Tkץ  */
static char **pargv;
Tk_ArgvInfo argTable[] = {
     {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
                      (char *) NULL }
};

#define F_MAX 7			/*      70 */
#define F_L6  6			/* 80 60 */
#define F_L5  5			/* 70 50 */
#define F_L4  4			/* 60 40 */
#define F_L3  3			/* 50 30 */
#define F_L2  2			/* 40 20 */
#define F_L1  1			/* 30 10 */
#define F_MIN 0			/* 20      */

void ClockSelecting(double f)
{
  static int downstat_time = 0;
  int mhz;
  int preclock_level, newclock_level;

  preclock_level = nowclock_level;

  newclock_level = F_MIN;
  if (f > 0.2) newclock_level = F_L1;
  if (f > 0.3) newclock_level = F_L2;
  if (f > 0.4) newclock_level = F_L3;
  if (f > 0.5) newclock_level = F_L4;
  if (f > 0.6) newclock_level = F_L5;
  if (f > 0.7) newclock_level = F_L6;
  if (f > 0.8) newclock_level = F_MAX;

  if (newclock_level >= nowclock_level) {
    /* Ψ徺ν(¨¤ȿ) */
    downstat_time = 0;
    nowclock_level = newclock_level;
  }
  else {
    /* Ψ߻ν(ø˲) */
    newclock_level = F_MIN;
    if (f > 0.1) newclock_level = F_L1;
    if (f > 0.2) newclock_level = F_L2;
    if (f > 0.3) newclock_level = F_L3;
    if (f > 0.4) newclock_level = F_L4;
    if (f > 0.5) newclock_level = F_L5;
    if (f > 0.6) newclock_level = F_L6;
    if (f > 0.7) newclock_level = F_MAX;

    if (newclock_level < nowclock_level && nowclock_level > 0) {
      downstat_time ++;
      if (downstat_time > 5) {
	nowclock_level --;
	downstat_time = 0;
      }
    }
    else
      downstat_time = 0;
  }

#if 0
  printf("[%f[%d,%d]]\n",f,newclock_level,FIDtable[newclock_level]);
  printf("[%f[%d,%d]]\n",f,nowclock_level,FIDtable[nowclock_level]);
#endif

  if (nowclock_level != preclock_level) {
    k7SetFID(FIDtable[nowclock_level],0);
  }
}

int observation(ClientData a, Tcl_Interp *interp, int argc, char **argv )
{
  char p[640];
  char *pp;
  int i, len, t;
  double f;
  int mhz;
#ifndef NO_SENSOR
  double Vcore, CPUfan;
#endif

  /*****************************************************************
   * 1å󥿡Х 
   */
  TEXEC( Tcl_Eval( interp, "after 1000 observation" ), "observation,");

  /*****************************************************************
   * CPUβƯΨΤäCPUå
   */
  f = cpustat();
  if (autof)
    ClockSelecting(f);


#ifndef NO_SENSOR
  /*****************************************************************
   * 󥵡
   */
  Vcore  = SensorGetVcore();
  CPUfan = SensorGetCPUfan();
  sprintf(p, ".control.vcore.value configure -text  \"%1.3fV\"",Vcore);
  TEXEC( Tcl_Eval( interp, p), "rate");
  sprintf(p, ".control.vcore.title configure -text  \"%2.1fC\"",SensorGetSYStemp());
  TEXEC( Tcl_Eval( interp, p), "rate");
  sprintf(p, ".control.fcpu.title  configure -text  \"%2.1fC\"",SensorGetCPUtemp());
  TEXEC( Tcl_Eval( interp, p), "rate");
  sprintf(p, ".control.cpufan.title configure -text  \"%3.0frpm\"",SensorGetSYSfan());
  TEXEC( Tcl_Eval( interp, p), "rate");
  sprintf(p, ".control.cpufan.value configure -text  \"%3.0frpm\"",CPUfan);
  TEXEC( Tcl_Eval( interp, p), "rate");

  sprintf(p, ".control.vcore.scale set %d",(int)(Vcore*100.0));
  TEXEC( Tcl_Eval(interp, p),"Vcore");
  sprintf(p, ".control.cpufan.scale set %d",(int)CPUfan);
  TEXEC( Tcl_Eval(interp, p),"CPUfan");
#endif

  /*****************************************************************
   * CPUβƯΨΥ
   */
  if (f >= 0) {
    sprintf(p, ".data.rate configure -text \"%3.1f%%\"",f*100);
    TEXEC( Tcl_Eval( interp, p), "rate");
    cpustates[time_index] = f;
  }
  /* printf("%f\n",f);		/*  */

  /* CPU user + system rate graph */
  pp = &p[0];
  sprintf(pp, ".data.canvas coords cpurate 199 19 ");
  t = time_index;
  for (i = 0; i < STATMAXNUM; i ++) {
    len = strlen(p);
    pp = &p[0]+len;
    sprintf(pp, "%d %d ",
	    119 - i*2,
	    19 - (int)(cpustates[t]*19.0));
    t --;
    if (t < 0) {
      t = STATMAXNUM - 1;
    }
  }
  len = strlen(p);
  pp = &p[0]+len;
  sprintf(pp, "%d %d", 0 , 19);
  TEXEC( Tcl_Eval( interp, p), "poly");

  /*****************************************************************
   * CPUβƯΨΤäCPUå
   */
  if (autof)
    ClockSelecting(f);

  /*****************************************************************
   * ưåΥ
   */
  mhz = k7GetClock(0);		/* ưȿ*/
  sprintf(p, ".control.fcpu.value configure -text \"%dMHz\"",mhz);
  TEXEC( Tcl_Eval( interp, p), "rate");

  sprintf(p, ".control.fcpu.scale set %d",mhz);
  TEXEC( Tcl_Eval(interp, p),"MHz,");
#if 0
  printf("<< %s >>",p);
#endif

  cpuclocks[time_index] = (double)(mhz - f_minimam)/(f_maximam - f_minimam);
  /* CPU Clock graph */
  pp = &p[0];
  sprintf(pp, ".data.canvas coords cpuclock ");
  t = time_index;
  for (i = 0; i < STATMAXNUM; i ++) {
    len = strlen(p);
    pp = &p[0]+len;
    sprintf(pp, "%d %d ", 
	    119 - i*2, 
	    19 - (int)(cpuclocks[t]*19));
    t --;
    if (t < 0) {
      t = STATMAXNUM - 1;
    }
  }
  TEXEC( Tcl_Eval( interp, p), "line");

  time_index++;
  if (time_index >= STATMAXNUM)
    time_index = 0;
  return TCL_OK;
}

int setclock(ClientData a, Tcl_Interp *interp, int argc, char **argv )
{
  int mhz,fid;
  char p[640];

  if (!autof) {
    mhz = atoi(argv[argc-1]);
    /*    clock = k7mkClock(mhz); */
    k7SetClock(mhz,0);
  }

  fid = k7GetFID(0);
  for (nowclock_level = 0 ;nowclock_level <8 ; nowclock_level ++) {
    if (fid == FIDtable[nowclock_level])
      break;
  }

  mhz = k7GetClock(0);		/* ưȿ*/
  sprintf(p, ".control.fcpu.scale set %d",mhz);
  TEXEC( Tcl_Eval(interp, p),"setclock,");

  return TCL_OK;
}

int auto_manual(ClientData a, Tcl_Interp *interp, int argc, char **argv )
{
  if (autof) {
    autof = 0;
    TEXEC( Tcl_Eval(interp, ".data.auto config -text \"manual\""),"auto_maual,");
  }
  else {
    autof = 1;
    TEXEC( Tcl_Eval(interp, ".data.auto config -text \" auto \""),"auto_maual,");
  }

  return TCL_OK;
}


void DisplayConfFile(char *str) 
{
/*
 * Init script
 */
  static char init_buffer[2048];

  f_minimam = ((float)FID_DATA[FIDtable[0]]*fsb);
  f_maximam = ((float)FID_DATA[FIDtable[7]]*fsb + 0.9);
  /*  printf("%d %d %f %f\n",f_minimam, f_maximam,((float)FID_DATA[FIDtable[7]]*fsb + 0.5),fsb); */

#ifndef NO_SENSOR
sprintf(init_buffer,"\n\
wm title . \"tkK7mauto\"\n\
wm iconname . \"tk7mauto\"\n\
after 1 { observation }\n\
frame .control\n\
frame .data\n\
pack .data .control -side top -expand yes -anchor s -fill x  -padx 5\n\
\n\
frame .control.vcore\n\
frame .control.fcpu\n\
frame .control.cpufan\n\
pack .control.fcpu .control.vcore .control.cpufan \
    -side top -expand yes -anchor s -fill x  -padx 5\n\
\n\
scale .control.vcore.scale -orient horizontal -length 120 -from 110 -to 185 \
 -font {fixed 8}\n\
label .control.vcore.title -text \"33.0C\" -font {fixed 10}\n\
label .control.vcore.value -text \"1.450V\" -font {fixed 10}\n\
pack .control.vcore.title .control.vcore.scale .control.vcore.value \
     -side left -expand 1\n\
\n\
scale .control.fcpu.scale -orient horizontal -length 120 -from %d -to %d \
     -command \"setclock .control.fcpu.scale\" -font {fixed 8}\n\
label .control.fcpu.title -text \"29.5C\" -font {fixed 10}\n\
label .control.fcpu.value -text \"600MHz\" -font {fixed 10}\n\
pack .control.fcpu.title .control.fcpu.scale .control.fcpu.value \
     -side left -expand 1\n\
\n\
scale .control.cpufan.scale -orient horizontal -length 120 -from 1000 -to 3000\
 -font {fixed 8}\n\
label .control.cpufan.title -text \"1170rpm\" -font {fixed 10}\n\
label .control.cpufan.value -text \"1200rpm\" -font {fixed 10}\n\
pack .control.cpufan.title .control.cpufan.scale .control.cpufan.value \
     -side left -expand 1\n\
\n\
button .data.auto -text \" auto \" -command {auto_manual} -font {fixed 10}\n\
canvas .data.canvas -width 120 -height 21 -bd 0 -highlightthickness 0\n\
.data.canvas create polygon 0 0 199 0 199 19 0 19 -fill white -tags canvas\n\
.data.canvas create polygon 0 0 1 1 2 2 -fill green -tags cpurate\n\
.data.canvas create line 0 0 1 1 1 2 0 0 -fill red -tags cpuclock\n\
label .data.rate  -text \"0.0%\"  -font {fixed 10}\n\
pack .data.auto .data.canvas .data.rate -side left -expand 1\n\
bind . <Control-q> \"exit\"\n\
\n\
.control.vcore.scale set 145\n\
.control.fcpu.scale set 600\n\
.control.cpufan.scale set 1500\n\
",f_minimam, f_maximam);
#else
/* NO_SENSOR */
sprintf(init_buffer,"\n\
wm title . \"tkK7mauto\"\n\
wm iconname . \"tk7mauto\"\n\
after 1 { observation }\n\
frame .control\n\
frame .data\n\
pack .control .data -side top -expand yes -anchor s -fill x  -padx 5\n\
\n\
frame .control.fcpu\n\
pack .control.fcpu  -side top -expand yes -anchor s -fill x  -padx 5\n\
\n\
scale .control.fcpu.scale -orient horizontal -length 120 -from %d -to %d \
     -command \"setclock .control.fcpu.scale\" -font {fixed 8}\n\
label .control.fcpu.title -text \"CPU\" -font {fixed 10}\n\
label .control.fcpu.value -text \"600MHz\" -font {fixed 10}\n\
pack .control.fcpu.title .control.fcpu.scale .control.fcpu.value \
     -side left -expand 1\n\
\n\
button .data.auto -text \" auto \" -command {auto_manual} -font {fixed 10}\n\
canvas .data.canvas -width 120 -height 21 -bd 0 -highlightthickness 0\n\
.data.canvas create polygon 0 0 199 0 199 19 0 19 -fill white -tags canvas\n\
.data.canvas create polygon 0 0 1 1 2 2 -fill green -tags cpurate\n\
.data.canvas create line 0 0 1 1 1 2 0 0 -fill red -tags cpuclock\n\
label .data.rate  -text \"0.0%\"  -font {fixed 10}\n\
pack .data.auto .data.canvas .data.rate -side left -expand 1\n\
bind . <Control-q> \"exit\"\n\
\n\
.control.fcpu.scale set 600\n\
",f_minimam, f_maximam);
#endif

 mix_init = init_buffer ;
}

int MyAppInit( Tcl_Interp *interp) 
{
   int argc = 0;
   char p[100];

   DisplayConfFile("dmts.cnf");

   for (; ( pargv[argc] ); argc++ );

   TEXEC( Tk_ParseArgv(interp, Tk_MainWindow( interp ), &argc, pargv, argTable, 0), "" );

   // init tcl/tk   
   TEXEC( Tcl_Init( interp ), "Tcl_Init failed" );
   TEXEC( Tk_Init( interp ),  "Tk_Init failed"  );
   
   // command 
   TEXEC2( Tcl_CreateCommand( interp, "observation", observation, NULL, NULL ),
	   "Create command failed" );
   TEXEC2( Tcl_CreateCommand( interp, "setclock", setclock, NULL, NULL ),
	   "Create command failed" );
   TEXEC2( Tcl_CreateCommand( interp, "auto_manual", auto_manual, NULL, NULL ),
	   "Create command failed" );
   
   // init GUI
   TEXEC( Tcl_Eval( interp, mix_init ), "Init failed" );

   // run tk engine   
   Tk_MainLoop();
   
   // for good feeling...   
   TEXEC( Tcl_Eval( interp, "exit 0" ), "exit failed" );
   
   return TCL_OK;
}

static struct ARGOPTION opt[] = { 
  { "-max",    INT,   0},	/* max cpu clock */
  { "",           END,    0}, 
};

int main( int argc, char **argv ) {
  int i,fid;
  pargv = argv;
  time_index = 0;
  nowclock_level = 0;

  f_minimam = 300;
  f_maximam = 2000;

  ArgGetOption(opt, argc, argv); 
  i = 0;
  while (opt[i].type != END ) {
    if (!strcmp("-max",opt[i].option)) { 
      if (opt[i].data.idata > 0) 
	f_maximam = (opt[i].data.idata); 
    } 
    i ++; 
  }

  fid = k7GetFID(0);
  for (nowclock_level = 0 ;nowclock_level <8 ; nowclock_level ++) {
    if (fid == FIDtable[nowclock_level])
      break;
  }

  fsb = makeFIDtable(8, (float)f_maximam, FIDtable);
  k7SetFIDTable(FIDtable,fsb);

#if 0
  printf("%f\n",fsb);
#endif

  for (i = 0; i < STATMAXNUM; i ++) {
    cpustates[i]= 0;
    cpuclocks[i]= 0;
  }
  autof = 1;

  Tk_Main( argc, argv, MyAppInit);
  return 0;
}
